package net.sf.sidaof.dao.hibernate;

import java.io.Serializable;
import java.util.List;

import net.sf.sidaof.dao.AbstractDao;

import org.hibernate.HibernateException;
import org.hibernate.LockMode;
import org.hibernate.NonUniqueResultException;
import org.hibernate.ObjectNotFoundException;
import org.hibernate.Query;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public abstract class AbstractHibernateDao<K extends Serializable, E> extends
        AbstractDao<K, E> implements HibernateDao<K, E> {
    private final Logger LOG =
            LoggerFactory.getLogger(AbstractHibernateDao.class);

    /** Make a new instance, automatically determining the entity class. */
    public AbstractHibernateDao() {
        super();
    }

    /** Make a new instance, specifying the entity class. */
    public AbstractHibernateDao(Class<E> entityClass) {
        super(entityClass);
    }

    @Override
    public void clear() {
        LOG.debug("clear: called");
        getSession().clear();
    }

    @Override
    public boolean contains(E entity) {
        LOG.debug("contains: entity={}", entity);
        boolean contains = getSession().contains(entity);
        return contains;
    }

    @Override
    @SuppressWarnings("unchecked")
    public List<E> findAll() {
        Query query = getSession().createQuery("from " + entityClass.getName());
        LOG.debug("findAll: query={}", query);
        List<E> list = query.list();
        return list;
    }

    @Override
    @SuppressWarnings("unchecked")
    public List<E> findAll(int startPosition, int maxResult) {
        LOG.debug("findAll: startPosition={}, maxResult={}", startPosition,
                maxResult);
        Query query =
                getSession().createQuery("from " + entityClass.getName())
                        .setFirstResult(startPosition).setMaxResults(maxResult);
        LOG.debug("findAll: query={}", query);
        List<E> list = query.list();
        return list;
    }

    @Override
    public E findById(K id) {
        LOG.debug("findById: id={}", id);
        return findById(id, LockMode.NONE, false);
    }

    @SuppressWarnings("unchecked")
    // TODO test what happens with null lockMode
    public E findById(K id, LockMode lockMode, boolean exceptionIfNotFound) {
        final Object[] logArgs = {id, lockMode, exceptionIfNotFound};
        LOG.debug("findById: id={}, lock={}, exceptionIfNotFound={}", logArgs);
        E entity = (E) getSession().get(entityClass, id, lockMode);
        if (entity == null) {
            final String msg =
                    "Entity " + entityClass.getName() + " with id " + id
                            + " not found";
            LOG.info("getById: {}", msg);
            if (exceptionIfNotFound) {
                throw new ObjectNotFoundException(id, entityClass.toString());
            }
        } else if (lockMode != null) {
            // don't try lock if entity is null
            getSession().lock(entity, lockMode);
        }
        return entity;
    }

    @Override
    public void flush() {
        LOG.debug("flush: called");
        getSession().flush();
    }

    @Override
    @SuppressWarnings("unchecked")
    public E merge(E entity) {
        LOG.debug("merge: entity={}", entity);
        E newEntity = (E) getSession().merge(entity);
        LOG.debug("merge: merged entity={}", newEntity);
        return newEntity;
    }

    @Override
    public void persist(E entity) {
        LOG.debug("persist: entity={}", entity);
        getSession().persist(entity);
        LOG.debug("persist: persisted entity={}", entity);
    }

    @Override
    public void refresh(E entity) {
        LOG.debug("refresh: old entity={}", entity);
        getSession().refresh(entity);
        LOG.debug("refresh: new entity={}", entity);
    }

    @Override
    public void remove(E entity) {
        LOG.debug("remove: entity={}", entity);
        getSession().delete(entity);
    }

    @Override
    public Long countAll(String where, String pathExpression) {
        if (where == null) {
            where = "";
        } else {
            where = " where " + where;
        }
        if (pathExpression == null) {
            pathExpression = "*";
        }

        final StringBuilder sb = new StringBuilder(1000);
        sb.append("select count(").append(pathExpression).append(") from ");
        sb.append(getEntityClass().getName());
        sb.append(" e");
        sb.append(where);

        String queryString = sb.toString();
        LOG.debug("countAll: queryString={}", queryString);

        Query q = getSession().createQuery(queryString);
        Long count = (Long) q.uniqueResult();
        return count;
    }

    @Override
    public E runQueryWithSingleResult(Query q) {
        LOG.debug("runQueryWithSingleResult: query={}", q);
        return runQueryWithSingleResult(q, null, false);
    }

    @Override
    @SuppressWarnings("unchecked")
    public E runQueryWithSingleResult(Query q, LockMode lockMode,
            boolean exceptionIfNotFound) {
        final Object[] logArgs = {q, lockMode, exceptionIfNotFound};
        LOG.debug("runQueryWithSingleResult:"
                + " query={}, lock={}, exceptionIfNotFound={}", logArgs);

        E entity = null;
        try {
            entity = (E) q.uniqueResult();
            if (entity == null) {
                // TODO how to use SLF4J to format the exception msg too?
                final String exMsg =
                        "runQueryWithSingleResult: found multiple of entity '"
                                + entityClass.getName() + "', query='" + q;

                // TODO what is a better exception than this? can't find one
                HibernateException e = new HibernateException(exMsg);
                handleQueryException(QUERY_SINGLERESULT_NOTFOUND_LOGMSG, e, q,
                        exceptionIfNotFound);
            } else if (lockMode != null) {
                getSession().lock(entity, lockMode);
            }
        } catch (NonUniqueResultException e) {
            final String msg =
                    QUERY_SINGLERESULT_NOTFOUND_LOGMSG + ", namedParams={}";
            handleQueryException(msg, e, q, exceptionIfNotFound);
        }
        return entity;
    }

    private void handleQueryException(String msg, HibernateException e,
            Query q, boolean exceptionIfNotFound) {
        Object[] logArgs = {entityClass.getName(), q, e};
        LOG.warn(msg, logArgs);
        if (exceptionIfNotFound) {
            throw e;
        }
    }
}
