package net.sf.sidaof.spring;

import javax.persistence.Entity;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.ClassPathScanningCandidateComponentProvider;
import org.springframework.core.type.filter.AnnotationTypeFilter;
import org.springframework.orm.jpa.persistenceunit.MutablePersistenceUnitInfo;
import org.springframework.orm.jpa.persistenceunit.PersistenceUnitPostProcessor;

/**
 * JPA Persistence Unit Spring Post Processor to scan for @Entity annotations,
 * as JPA scanning isn't very thorough.
 * 
 * Configuration Example: Add a persistenceUnitPostProcessors definition to the
 * entityManagerFactory bean definition, such as:
 * 
 * <pre>
 * &lt;property name=&quot;persistenceUnitPostProcessors&quot;&gt;
 *   &lt;bean class=&quot;net.sf.sidaof.spring.EntityPersistenceUnitPostProcessor&quot;&gt;
 *     &lt;property name=&quot;packages&quot; value=&quot;com.mypackage.domain&quot;&gt;&lt;/property&gt;
 *   &lt;/bean&gt;
 * &lt;/property&gt;
 * </pre>
 */
public class EntityPersistenceUnitPostProcessor implements
        PersistenceUnitPostProcessor, InitializingBean {
    private static final Logger LOG =
            LoggerFactory.getLogger(EntityPersistenceUnitPostProcessor.class);

    /** The packages to scan for. */
    private String[] packages = new String[] {};

    /** The classpath scanner. */
    private ClassPathScanningCandidateComponentProvider provider;

    /**
     * {@inheritDoc}
     * 
     * Search for persistent classes under the configured folders and register
     * them in the persistence unit.
     */
    @Override
    public void postProcessPersistenceUnitInfo(MutablePersistenceUnitInfo pui) {
        for (String pkg : packages) {
            for (BeanDefinition bean : provider.findCandidateComponents(pkg)) {
                final String className = bean.getBeanClassName();
                LOG.debug("adding entity={}", className);
                pui.addManagedClassName(className);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void afterPropertiesSet() throws Exception {
        if (provider == null) {
            provider = new ClassPathScanningCandidateComponentProvider(false);
            provider.addIncludeFilter(new AnnotationTypeFilter(Entity.class,
                    false));
        }
    }

    /**
     * Get the provider.
     * 
     * @see {@link provider}.
     * 
     * @return The provider.
     */
    public ClassPathScanningCandidateComponentProvider getProvider() {
        return provider;
    }

    /**
     * Set the provider.
     * 
     * @see {@link provider}.
     * 
     * @param provider
     *            The provider to set.
     */
    public void setProvider(ClassPathScanningCandidateComponentProvider provider) {
        this.provider = provider;
    }

    /**
     * Get the packages.
     * 
     * @see {@link packages}.
     * 
     * @return The packages.
     */
    public String[] getPackages() {
        return packages;
    }

    /**
     * Set the packages.
     * 
     * @see {@link packages}.
     * 
     * @param packages
     *            The packages to set.
     */
    public void setPackages(String[] packages) {
        this.packages = packages;
    }
}
